Package: projpred
Encoding: UTF-8
Title: Projection Predictive Feature Selection
Version: 2.2.0
Date: 2022-08-18
Authors@R: c(person("Juho", "Piironen", role = c("aut"),
                    email = "juho.t.piironen@gmail.com"),
             person("Markus", "Paasiniemi", role = "aut"),
             person("Alejandro", "Catalina", role = "aut",
                    email = "alecatfel@gmail.com"),
             person("Frank", "Weber", role = c("cre", "aut"),
                    email = "fweber144@protonmail.com"),
             person("Aki", "Vehtari", role = "aut"),
             person("Jonah", "Gabry", role = "ctb"),
             person("Marco", "Colombo", role = "ctb"),
             person("Paul-Christian", "Bürkner", role = "ctb"),
             person("Hamada S.", "Badr", role = "ctb"))
Description: 
    Performs projection predictive feature selection for generalized linear and
    additive models as well as for generalized linear and additive multilevel
    models (see Piironen, Paasiniemi and Vehtari, 2020,
    <doi:10.1214/20-EJS1711>; Catalina, Bürkner and Vehtari, 2020,
    <arXiv:2010.06994>). The package is compatible with the 'rstanarm' and
    'brms' packages, but other reference models can also be used. See the
    documentation as well as the package vignette for more information and
    examples.
License: GPL-3 | file LICENSE
URL: https://mc-stan.org/projpred/, https://discourse.mc-stan.org
BugReports: https://github.com/stan-dev/projpred/issues/
Depends: R (>= 3.5.0)
Imports: methods, dplyr, loo (>= 2.0.0), rstantools (>= 2.0.0), lme4,
        mvtnorm, ggplot2, Rcpp, utils, magrittr, mgcv, gamm4, rlang
Suggests: rstanarm, brms, testthat, knitr, rmarkdown, glmnet, cmdstanr,
        bayesplot (>= 1.5.0), optimx, posterior, parallel, foreach,
        iterators, doParallel, future, future.callr, doFuture
LinkingTo: Rcpp, RcppArmadillo
Additional_repositories: https://mc-stan.org/r-packages/
LazyData: TRUE
RoxygenNote: 7.2.1
VignetteBuilder: knitr, rmarkdown
NeedsCompilation: yes
Packaged: 2022-08-19 07:32:07 UTC; frank
Author: Juho Piironen [aut],
  Markus Paasiniemi [aut],
  Alejandro Catalina [aut],
  Frank Weber [cre, aut],
  Aki Vehtari [aut],
  Jonah Gabry [ctb],
  Marco Colombo [ctb],
  Paul-Christian Bürkner [ctb],
  Hamada S. Badr [ctb]
Maintainer: Frank Weber <fweber144@protonmail.com>
Repository: CRAN
Date/Publication: 2022-08-19 08:40:02 UTC
