Package: projpred
Encoding: UTF-8
Title: Projection Predictive Feature Selection
Version: 2.10.0
Date: 2025-12-05
Authors@R: c(person("Juho", "Piironen", role = c("aut"),
                    email = "juho.t.piironen@gmail.com"),
             person("Markus", "Paasiniemi", role = "aut"),
             person("Alejandro", "Catalina", role = "aut",
                    email = "alecatfel@gmail.com"),
             person("Frank", "Weber", role = "aut"),
             person("Osvaldo", "Martin", role = c("cre", "aut"),
                    email = "aloctavodia@gmail.com"),
             person("Aki", "Vehtari", role = "aut"),
             person("Jonah", "Gabry", role = "ctb"),
             person("Marco", "Colombo", role = "ctb"),
             person("Paul-Christian", "Bürkner", role = "ctb"),
             person("Hamada S.", "Badr", role = "ctb"),
             person("Brian", "Sullivan", role = "ctb"),
             person("Sölvi", "Rögnvaldsson", role = "ctb"),
             person(family = "The LME4 Authors", role = "cph",
                    comment = "see file 'LICENSE' for details"),
             person("Yann", "McLatchie", role = "ctb"),
             person("Juho", "Timonen", role = "ctb"))
Description: 
    Performs projection predictive feature selection for generalized linear
    models (Piironen, Paasiniemi, and Vehtari, 2020, <doi:10.1214/20-EJS1711>)
    with or without multilevel or additive terms (Catalina, Bürkner, and
    Vehtari, 2022, <https://proceedings.mlr.press/v151/catalina22a.html>), for
    some ordinal and nominal regression models (Weber, Glass, and Vehtari, 2025,
    <doi:10.1007/s00180-024-01506-0>), and for many other regression models
    (using the latent projection by Catalina, Bürkner, and Vehtari, 2021,
    <doi:10.48550/arXiv.2109.04702>, which can also be applied to most of the
    former models). The package is compatible with the 'rstanarm' and 'brms'
    packages, but other reference models can also be used. See the vignettes and
    the documentation for more information and examples.
License: GPL-3 | file LICENSE
URL: https://mc-stan.org/projpred/, https://discourse.mc-stan.org
BugReports: https://github.com/stan-dev/projpred/issues/
Depends: R (>= 3.6.0)
Imports: methods, utils, Rcpp, gtools, ggplot2, scales, rstantools (>=
        2.0.0), loo (>= 2.0.0), lme4 (>= 1.1-28), mvtnorm, mgcv, gamm4,
        abind, MASS, ordinal, nnet, mclogit, reformulas,
Suggests: ggrepel, ggfortify, rstanarm, brms, nlme, optimx, ucminf,
        parallel, foreach, iterators, doRNG, unix, testthat, vdiffr,
        knitr, rmarkdown, glmnet, cmdstanr, rlang, bayesplot (>=
        1.5.0), posterior, doParallel, future, future.callr, doFuture,
        progressr
LinkingTo: Rcpp, RcppArmadillo
Additional_repositories: https://stan-dev.r-universe.dev/
LazyData: TRUE
RoxygenNote: 7.3.3
VignetteBuilder: knitr, rmarkdown
NeedsCompilation: yes
Packaged: 2025-12-05 16:04:36 UTC; osvaldo
Author: Juho Piironen [aut],
  Markus Paasiniemi [aut],
  Alejandro Catalina [aut],
  Frank Weber [aut],
  Osvaldo Martin [cre, aut],
  Aki Vehtari [aut],
  Jonah Gabry [ctb],
  Marco Colombo [ctb],
  Paul-Christian Bürkner [ctb],
  Hamada S. Badr [ctb],
  Brian Sullivan [ctb],
  Sölvi Rögnvaldsson [ctb],
  The LME4 Authors [cph] (see file 'LICENSE' for details),
  Yann McLatchie [ctb],
  Juho Timonen [ctb]
Maintainer: Osvaldo Martin <aloctavodia@gmail.com>
Repository: CRAN
Date/Publication: 2025-12-06 13:30:02 UTC
