# Generated by roxygen2: do not edit by hand

S3method("[",augmat)
S3method("[",augvec)
S3method(as.matrix,projection)
S3method(coef,subfit)
S3method(cv_proportions,ranking)
S3method(cv_proportions,vsel)
S3method(cv_varsel,default)
S3method(cv_varsel,refmodel)
S3method(cv_varsel,vsel)
S3method(formula,gamm4)
S3method(get_refmodel,default)
S3method(get_refmodel,projection)
S3method(get_refmodel,refmodel)
S3method(get_refmodel,stanreg)
S3method(get_refmodel,vsel)
S3method(get_subparams,clmm)
S3method(get_subparams,gamm4)
S3method(get_subparams,glm)
S3method(get_subparams,glmerMod)
S3method(get_subparams,glmmPQL)
S3method(get_subparams,lm)
S3method(get_subparams,lmerMod)
S3method(get_subparams,mmblogit)
S3method(get_subparams,multinom)
S3method(get_subparams,polr)
S3method(get_subparams,subfit)
S3method(performances,vsel)
S3method(performances,vselsummary)
S3method(plot,cv_proportions)
S3method(plot,ranking)
S3method(plot,vsel)
S3method(posterior::as_draws,projection)
S3method(posterior::as_draws_matrix,projection)
S3method(predict,gamm4)
S3method(predict,refmodel)
S3method(predict,subfit)
S3method(predictor_terms,projection)
S3method(print,projection)
S3method(print,refmodel)
S3method(print,vsel)
S3method(print,vselsummary)
S3method(ranking,vsel)
S3method(repair_re,clmm)
S3method(repair_re,merMod)
S3method(repair_re,mmblogit)
S3method(run_cvfun,default)
S3method(run_cvfun,refmodel)
S3method(solution_terms,projection)
S3method(solution_terms,vsel)
S3method(suggest_size,vsel)
S3method(summary,vsel)
S3method(t,augmat)
S3method(t,augvec)
S3method(varsel,default)
S3method(varsel,refmodel)
S3method(varsel,vsel)
export(Student_t)
export(augdat_ilink_binom)
export(augdat_link_binom)
export(break_up_matrix_term)
export(cl_agg)
export(cv_folds)
export(cv_ids)
export(cv_proportions)
export(cv_varsel)
export(cvfolds)
export(do_call)
export(extend_family)
export(force_search_terms)
export(get_refmodel)
export(init_refmodel)
export(performances)
export(predictor_terms)
export(proj_linpred)
export(proj_predict)
export(project)
export(ranking)
export(run_cvfun)
export(solution_terms)
export(suggest_size)
export(varsel)
export(y_wobs_offs)
import(ggplot2)
import(stats)
importFrom(Rcpp,sourceCpp)
importFrom(loo,kfold)
importFrom(methods,as)
importFrom(methods,cbind2)
importFrom(methods,is)
importFrom(rstantools,posterior_linpred)
useDynLib(projpred)
