% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_formula.R
\name{helper_formula}
\alias{helper_formula}
\title{Utilities to handle formulas for the external user}
\description{
Utilities to handle formulas for the external user
}
